! This program solves the sudoku puzzle.
! The following kinds are supported:
! STD:  the classic sudoku
! WIN:  the sudoku as published by the NRC: 4 extra units of 3x3,
!       located with one row/column spacing from the outer edge
!       Also known as windoku. See https://www.sudokuwiki.org/windoku.aspx
! XXX:  aka sudoku-X: the diagonals also play a role
!

! the program expects the following input:
! - a string, WIN or STD or XXX to signify the type of sudoku
!   possibly followed by comment. Only the first 3 characters
!   of the first line are read.
! - 81 integers denoting the contents of the board, row-wise.
!   an empty cell is denoted by zero or '.'
! - other characters are ignored
! example:
!    WIN   an example from the NRC
!    3 4 0 0 6 0 0 0 8
!    1 0 0 0 8 0 0 0 0
!    0 2 0 0 0 5 0 0 0
!    0 0 0 0 0 0 0 9 2
!    0 0 5 0 0 9 0 3 7
!    8 0 7 0 4 0 0 0 0
!    0 0 0 0 0 3 0 0 5
!    0 1 0 0 0 0 0 0 0
!    0 0 0 0 9 0 0 0 0

module fsudoku_module
   implicit none
   integer, private                       :: nunits
   integer, target, private               :: a(9,9)
   logical, private                       :: aa(9,9)
   integer, allocatable, private          :: unit(:,:)
   integer, allocatable, private          :: inunits(:,:,:)
   type punit_type
      integer, pointer                    :: p(:,:)
   end type
   type(punit_type), allocatable, private :: punit(:)
   character(len=3),private               :: sudoku_type
   private                                :: has, possible
   integer, parameter, private            :: standard = 1
   integer, parameter, private            :: windoku  = 2
   integer, parameter, private            :: xsudoku  = 3
   integer, private                       :: sudoku_t

   integer                                :: number_of_solutions

contains
   integer function rightmostbit(i)
      integer, intent(in) :: i
      rightmostbit = iand(i,-i)
   end function

   integer function myand(a,b)
      integer, intent(in) :: a,b
      integer i,a1,b1,ab,bb,m
      a1 = a
      b1 = b
      myand = 0
      m = 1
      do  i=1,32
         if(a1 .eq. 0) return
         if(b1 .eq. 0) return
         ab = mod(a1,2)
         bb = mod(b1,2)
         !print *,i,a1,b1,ab,bb
         myand = myand + m*((abs(ab)+abs(bb))/2)
         m = 2*m
         a1 = a1/2
         b1 = b1/2
         if (ab .eq. -1) a1 = a1-1
         if (bb .eq. -1) b1 = b1-1
      enddo
   end function

   integer function rm1(i)
      integer, intent(in) :: i
      rm1 = myand(i,-i)
   end function

   subroutine sudoku_init
      integer :: i,j,k,p,maxp,i1,i2,j1,j2
      integer, allocatable :: intemp(:,:,:)

      number_of_solutions = 0
      select case (sudoku_type)
       case('WIN')
         nunits = 3*9+4
         sudoku_t = windoku
       case('XXX')
         nunits = 3*9
         sudoku_t = xsudoku
       case default
         nunits = 3*9
         sudoku_t = standard
       case('STD')
         nunits = 3*9
         sudoku_t = standard
      end select

      allocate(unit(4,nunits))

      ! unit will contain start and end coordinates of units
      ! unit(1,:) startrow
      ! unit(2,:) endrow
      ! unit(3,:) startcolumn
      ! unit(4,:) endcolumn

      k = 1
      do i=1,9 !rows
         unit(1,k) = i
         unit(2,k) = i
         unit(3,k) = 1
         unit(4,k) = 9
         k = k+1
      enddo
      do i=1,9 ! columns
         unit(1,k) = 1
         unit(2,k) = 9
         unit(3,k) = i
         unit(4,k) = i
         k = k+1
      enddo
      do i=1,3 ! squares
         do j=1,3
            unit(1,k) = (i-1)*3 + 1
            unit(2,k) = unit(1,k) + 2
            unit(3,k) = (j-1)*3 + 1
            unit(4,k) = unit(3,k) + 2
            k = k+1
         enddo
      enddo
      if (sudoku_t .eq. windoku) then
         do i=1,2 ! windoku squares
            do j=1,2
               unit(1,k) = (i-1)*4 + 2
               unit(2,k) = unit(1,k) + 2
               unit(3,k) = (j-1)*4 + 2
               unit(4,k) = unit(3,k) + 2
               k = k+1
            enddo
         enddo
      endif

      allocate(punit(nunits))
      do i=1,nunits
         i1 = unit(1,i)
         i2 = unit(2,i)
         j1 = unit(3,i)
         j2 = unit(4,i)
         punit(i)%p => a(i1:i2,j1:j2)
      enddo

      ! For all cells on board: in which unit are they.
      ! inunits(:,i,j) contains index numbers of units wherein
      ! the cell(i,j) resides in. A zero means end of the list

      allocate(intemp(nunits+1,9,9))
      intemp = 0
      maxp = 0
      do i=1,9
         do j=1,9
            p = 0
            do k=1,nunits
               if (i .ge. unit(1,k) .and.    &
                  i .le. unit(2,k) .and.     &
                  j .ge. unit(3,k) .and.     &
                  j .le. unit(4,k) ) then
                  p = p+1
                  intemp(p,i,j) = k
               endif
            enddo
            if (p .gt. maxp) maxp = p
         enddo
      enddo
      allocate (inunits(maxp,9,9))
      do i=1,9
         do j=1,9
            inunits(:,i,j) = intemp(1:maxp,i,j)
         enddo
      enddo
      deallocate(intemp)

   end subroutine sudoku_init

   ! pretty prints board on standard output
   ! invalid cells are colored red

   subroutine print_board
      integer :: i,j
      print *,sudoku_type
      do i=1,9
         do j=1,9
            if (mod(j-1,3) .eq.0) then
               write(*,'(a)',advance='no') ' |'
            endif
            if (a(i,j) .eq. 0) then
               write(*,'(a)',advance='no') ' .'
            else
               !aij = a(i,j)
               !a(i,j) = 0
               if (possible(i,j)) then
                  write(*,'(i2)',advance='no') a(i,j)
               else
                  if(command_argument_count() .eq. 0) then
                     write(*,'(a)',advance='no') char(27)//'[1;31m'
                     write(*,'(i2)',advance='no') a(i,j)
                     write(*,'(a)',advance='no') char(27)//'[0m'
                  else
                     write(*,'(i2)',advance='no') -a(i,j)
                  endif
               endif
               !a(i,j) = aij
            endif
         enddo
         write(*,'(a)') ' |'
         if (mod(i,3) .eq. 0) then
            write(*,*)
         endif
      enddo
   end subroutine print_board

   ! write board to standard output. This can be used to read in again

   subroutine write_board
      integer :: i
      print *,sudoku_type
      do i=1,9
         print *,a(i,:)
      enddo
   end subroutine write_board

   ! write board to file f. This can be used to read in again

   subroutine fwrite_board(f)
      character(len=*), intent(in) :: f
      integer :: i,ff
      open(newunit=ff,file=f)
      write(ff,*) sudoku_type
      do i=1,9
         write(ff,'(9i3)') a(i,:)
      enddo
      close(ff)
   end subroutine fwrite_board

   ! read board from standard input

   subroutine read_board
      integer  :: i,j
      character:: c

      read *,sudoku_type
      select case (sudoku_type)
       case ('STD')
         sudoku_t = standard
         continue
       case ('WIN')
         sudoku_t = windoku
         continue
       case ('XXX')
         sudoku_t = xsudoku
         continue
       case default
         sudoku_type = 'STD'
         sudoku_t = standard
      end select
      a = -1
      i = 1
      j = 1
      loop: do
         do
            read(*,'(a1)',end=1010, eor=1000, advance='no') c
1000        continue
            if (c .eq. '.') c='0'
            if (c .ge. '0' .and. c .le. '9') then
               a(i,j) = ichar(c)-ichar('0')
               j = j+1
               if (j .gt. 9) then
                  j = 1
                  i = i+1
                  if (i .gt. 9) exit loop
               endif
            endif
         enddo
      enddo loop
      return
1010  continue
      print *,'Missing value for row',i,'and column',j, 'exiting ...'
      stop
   end subroutine read_board

   ! read board from file f

   subroutine fread_board(f)
      character(len=*), intent(in) :: f
      integer  :: i,ff,j
      character:: c

      open(newunit=ff,file=f)
      read(ff,*) sudoku_type
      select case (sudoku_type)
       case ('STD')
         sudoku_t = standard
         continue
       case ('WIN')
         sudoku_t = windoku
         continue
       case ('XXX')
         sudoku_t = xsudoku
         continue
       case default
         sudoku_t = standard
         sudoku_type = 'STD'
      end select
      a = -1
      i = 1
      j = 1
      loop: do
         do
            read(ff,'(a1)',end=1030, eor=1020, advance='no') c
1020        continue
            if (c .ge. '0' .and. c .le. '9') then
               a(i,j) = ichar(c)-ichar('0')
               j = j+1
               if (j .gt. 9) then
                  j = 1
                  i = i+1
                  if (i .gt. 9) exit loop
               endif
            endif
         enddo
      enddo loop
      close(ff)
      return
1030  continue
      print *,'Missing value for row',i,'and column',j, 'exiting ...'
      close(ff)
      stop
   end subroutine fread_board

   ! check if c contains n
   !  Note: usage of any() or findloc() results in poorer performance

   pure logical function has(c,n) result(r)
      integer, intent(in) :: c(:,:), n
      integer :: i,j
      r = .true.
      do i=1,size(c,1)
         do j=1,size(c,2)
            if (c(i,j) .eq. n) return
         enddo
      enddo
      r = .false.
      return
   end function

   ! check if a(i,j) is possible

   logical function possible(i,j) result(r)
      integer, intent(in) :: i,j
      integer :: k,ii,aij

      r = .false.

      aij    = a(i,j)
      a(i,j) = 0

      ! 3x3 blocks and rows and columns:
      do k=1,size(inunits,1)
         ii = inunits(k,i,j)
         if (ii .eq. 0) then
            exit
         endif

         if (has(punit(ii)%p,aij)) then
            goto 1040
         endif
      enddo

      ! diagonals:
      if(sudoku_t .eq. xsudoku) then
         if (i .eq. j) then
            do ii=1,9
               if (aij .eq. a(ii,ii)) goto 1040
            enddo
         endif
         if (i .eq. 10-j) then
            do ii=1,9
               if (aij .eq. a(ii,10-ii)) goto 1040
            enddo
         endif
      endif
      r = .true.

1040  continue
      a(i,j) = aij
      return

   end function


   ! returns i,j of next cell that has to be determined
   subroutine nextcell(i,j)
      integer, intent(inout) :: i,j
      do
         j = j+1
         if (j .gt. 9) then
            i = i+1
            j = 1
         endif
         if (i .gt. 9) return
         if (.not. aa(i,j)) return
      enddo
   end subroutine

   ! returns i,j of previous cell that has to be determined
   subroutine prevcell(i,j)
      integer, intent(inout) :: i,j
      do
         j = j-1
         if (j .lt. 1) then
            i = i-1
            j = 9
         endif
         if (i .lt. 1) return
         if (.not. aa(i,j)) return
      enddo
   end subroutine

   subroutine epilogue(n)
      integer, intent(in) :: n
      if(command_argument_count() .ne. 0) return
      print *,"Number of solutions is",number_of_solutions
      if(number_of_solutions .ge. n) then
         print *,"Probably, there are more solutions"
      endif
      stop
   end subroutine

   subroutine handle_solution(n)
      integer, intent(in) :: n
      number_of_solutions = number_of_solutions + 1
      call sudoku_solution
      if(command_argument_count() .ne. 0) then
         if(number_of_solutions .eq. 1) then
            print *,"Solution:"
            call print_board
         endif
         if(number_of_solutions .eq. 2) then
            print *,"**********************************"
            print *,"* This is NOT the only solution. *"
            print *,"**********************************"
            stop
         endif
      else
         print *,"Solution nr",number_of_solutions
         call print_board
         if (number_of_solutions .ge. n) call epilogue(n)
      endif
   end subroutine

   ! solves sudoku, at most nmax solutions are printed

   subroutine sudoku_solver(nmax)
      integer, intent(in) :: nmax
      integer :: i,j

      aa = a .ne. 0  ! to remember which positions are fixed

      i = 0
      j = 9

      ! nextcell will point (i,j) to the first modifiable cell
      call nextcell(i,j)

      ! If (i,j) is outside board, we have an already solved sudoku:
      if (i .gt. 9) then
         call handle_solution(nmax)
         return
      endif

      do

         ! Invariant: at this point, i,j is always the cell under investigation

         ! Try the next value for a(i,j):
         a(i,j) = a(i,j) + 1
         if (a(i,j) .gt. 9) then
            ! No solution given the previous cells: clear and step back
            a(i,j) = 0
            call prevcell(i,j)
            ! If we leave the board, there is nothing more to investigate
            if (i .lt. 1) exit
            cycle
         endif

         ! If a(i,j) has a possible value, go to the next cell
         if (possible(i,j)) then
            call nextcell(i,j)
            if (i .gt. 9) then
               ! Stepped outside board, so we have a solution
               call handle_solution(nmax)
               ! prevcell will let (i,j) point to the last modifiable cell
               call prevcell(i,j)
            endif
         endif

      enddo

   end subroutine

   ! returns .true. if this is a valid sudoku

   logical function sudoku_correct() result(r)
      integer :: i,j
      r = .true.
      do i=1,9
         do j=1,9
            if(a(i,j) .lt. 0 .or. a(i,j) .gt. 9) then
               r = .false.
               return
            endif
            if (a(i,j) .ne. 0) then
               if (.not. possible(i,j)) then
                  r = .false.
                  return
               endif
            endif
         enddo
      enddo
   end function


   subroutine sudoku_solution
      if (command_argument_count() .eq. 0) return
      call fwrite_board("sudoku-solution")
   end subroutine


end module fsudoku_module

program fsudoku
   use fsudoku_module
   integer :: nmax = 100
   if (command_argument_count() .ne. 0) nmax = 1

   call read_board()
   !call fread_board("std1.txt")
   call sudoku_init
   print *,"input:"
   call print_board
   if(.not. sudoku_correct()) then
      print *,'This is not a correct input.'
      return
   endif

   !call sudoku_solver(1,1,nmax)
   call sudoku_solver(nmax)

   call epilogue(nmax)

end program fsudoku
